import * as THREE from '../build/three.module.js';
import Stats from './jsm/libs/stats.module.js';
import { BufferGeometryUtils } from './jsm/utils/BufferGeometryUtils.js'

class App {
    constructor() {
        this._initialize();
    }

    _initialize() {
        this.domWebGL = document.createElement('div');
        document.body.appendChild(this.domWebGL);

        const renderer = new THREE.WebGLRenderer({ antialias: true });
        this.domWebGL.appendChild(renderer.domElement);
        this._renderer = renderer;

        const scene = new THREE.Scene();
        scene.background = new THREE.Color(0x050505);
        this._scene = scene;

        this._setupCamera();
        this._setupModel();
        this._setupGUI();
        
        window.onresize = this.resize.bind(this);

        this.resize();

        this.render();
    }

    _setupCamera() {
        const camera = new THREE.PerspectiveCamera(45, window.innerWidth / window.innerHeight, 1, 10000);
        camera.position.z = 300;
        this._camera = camera;
    }

    _createShaderMaterial() {
        const texture = new THREE.TextureLoader().load("textures/sprites/disc.png");
        texture.wrapS = THREE.RepeatWrapping;
        texture.wrapT = THREE.RepeatWrapping;

        const uniforms = {
            "color": { value: new THREE.Color(0xffffff) },
            pointTexture: { value: texture }
        };

        const vertexCode = `
            attribute float size;
            attribute vec3 ca; 
            
            varying vec3 vColor;

            void main() {
                vColor = ca;
                vec4 mvPosition = modelViewMatrix * vec4(position, 1.0);
                gl_PointSize = size * (300.0 / -mvPosition.z);
                gl_Position = projectionMatrix * mvPosition;
            }
        `;

        const fragmentCode = `
            uniform vec3 color;
            uniform sampler2D pointTexture;

            varying vec3 vColor;

            void main() {
                vec4 color = vec4(color * vColor, 1.0) * texture2D(pointTexture, gl_PointCoord);
                gl_FragColor = color;
            }
        `;
        

        const shaderMaterial = new THREE.ShaderMaterial({
            uniforms: uniforms,
            vertexShader: vertexCode,
            fragmentShader: fragmentCode,
            //blending: THREE.AdditiveBlending,
            //depthTest: false,
            transparent: true
        });

        return shaderMaterial;
    }

    _setupModel() {
        const radius = 100, segments = 68, rings = 38;
        let sphereGeometry = new THREE.SphereGeometry(radius, segments, rings);
        let boxGeometry = new THREE.BoxGeometry(0.8 * radius, 0.8 * radius, 0.8 * radius, 10, 10, 10);

        // if normal and uv attributes are not removed, mergeVertices() can't consolidate indentical vertices with different normal/uv data
        sphereGeometry.deleteAttribute("normal");
        sphereGeometry.deleteAttribute("uv");
        boxGeometry.deleteAttribute("normal");
        boxGeometry.deleteAttribute("uv");

        sphereGeometry = BufferGeometryUtils.mergeVertices(sphereGeometry);
        boxGeometry = BufferGeometryUtils.mergeVertices(boxGeometry);

        const combinedGeometry = BufferGeometryUtils.mergeBufferGeometries([ sphereGeometry, boxGeometry ]);
        const positionAttribute = combinedGeometry.getAttribute("position");

        const colors = [];
        const sizes = [];

        const color = new THREE.Color();
        const vertex = new THREE.Vector3();

        const length1 = sphereGeometry.getAttribute("position").count;
        for(let i=0, l=positionAttribute.count; i<l; i++) {
            vertex.fromBufferAttribute(positionAttribute, i);

            if(i < length1) {
                color.setHSL(0.01+0.1*(i/length1), 0.99, (vertex.y+radius)/(4*radius));
            } else {
                color.setHSL(0.6, 0.75, 0.25+vertex.y/(2*radius));
            }

            color.toArray(colors, i*3);
            sizes[i] = i < length1 ? 10 : 40;
        }

        const geometry = new THREE.BufferGeometry();
        geometry.setAttribute("position", positionAttribute);
        geometry.setAttribute("size", new THREE.Float32BufferAttribute(sizes, 1));
        geometry.setAttribute("ca", new THREE.Float32BufferAttribute(colors, 3));

        const sphere = new THREE.Points(geometry, this._createShaderMaterial());
        this._scene.add(sphere);

        this._sphere = sphere;
        this._length1 = length1;
    }

    _sortPoints() {
        const vector = new THREE.Vector3();

        // Model View Projection matrix
        const matrix = new THREE.Matrix4();
        matrix.multiplyMatrices(this._camera.projectionMatrix, this._camera.matrixWorldInverse);
        matrix.multiply(this._sphere.matrixWorld);
        // .

        const geometry = this._sphere.geometry;
        let index = geometry.getIndex();
        const positions = geometry.getAttribute("position").array;
        const length = positions.length / 3;

        if(index === null) {
            const array = new Uint16Array(length);
            for(let i=0; i<length; i++) {
                array[i] = i;
            }

            index = new THREE.BufferAttribute(array, 1);
            geometry.setIndex(index);
        }

        const sortArray = [];
        for(let i=0; i<length; i++) {
            vector.fromArray(positions, i*3);
            vector.applyMatrix4(matrix);
            sortArray.push([vector.z, i]);
        }

        sortArray.sort((a, b) => {
            return b[0] - a[0];
        });

        const indices = index.array;
        for(let i=0; i<length; i++) {
            indices[i] = sortArray[i][1];
        }

        geometry.index.needsUpdate = true;
    }

    update() {
        const time = Date.now() * 0.005;

        this._sphere.rotation.y = 0.02 * time;
        this._sphere.rotation.z = 0.02 * time;

        const geometry = this._sphere.geometry;
        const attributes = geometry.attributes;

        for(let i=0; i<attributes.size.array.length; i++) {
            if(i < this._length1) {
                attributes.size.array[i] = 16 + 12 * Math.sin(0.1 * i + time);
            }
        }

        attributes.size.needsUpdate = true;        

        this._sortPoints();
    }

    render() {
        requestAnimationFrame(this.render.bind(this));
        
        this.update();
        this._renderer.render(this._scene, this._camera);
        this._stats.update();
    }

    resize() {
        this._camera.aspect = window.innerWidth / window.innerHeight;
        this._camera.updateProjectionMatrix();

        this._renderer.setPixelRatio(window.devicePixelRatio);
        this._renderer.setSize(window.innerWidth, window.innerHeight);
    }

    _setupGUI() {
        const stats = new Stats();
        this.domWebGL.appendChild(stats.dom);
        this._stats = stats;
    }
}

window.onload = function () {
    (new App()).render();
}